// Copyright 2019 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:async';

import '../../firebase_auth_platform_interface.dart';
import '../firebase_auth_exception.dart';
import 'method_channel_user.dart';
import '../platform_interface/platform_interface_user_credential.dart';
import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

import 'method_channel_user_credential.dart';
import 'utils/exception.dart';
import 'utils/phone_auth_callbacks.dart';

/// Method Channel delegate for [FirebaseAuthPlatform].
class MethodChannelFirebaseAuth extends FirebaseAuthPlatform {
  /// Keep an internal reference to whether the [MethodChannelFirebaseAuth]
  /// class has already been initialized.
  static bool _initialized = false;

  /// Keeps an internal handle ID for the channel.
  static int _methodChannelHandleId = 0;

  /// Increments and returns the next channel ID handler for Auth.
  static int get nextMethodChannelHandleId => _methodChannelHandleId++;

  /// The [MethodChannelFirebaseAuth] method channel.
  static const MethodChannel channel = MethodChannel(
    'plugins.flutter.io/firebase_auth',
  );

  static Map<String, MethodChannelFirebaseAuth>
      _methodChannelFirebaseAuthInstances =
      <String, MethodChannelFirebaseAuth>{};

  static Map<String, StreamController<UserPlatform?>>
      _authStateChangesListeners = <String, StreamController<UserPlatform?>>{};

  static Map<String, StreamController<UserPlatform?>> _idTokenChangesListeners =
      <String, StreamController<UserPlatform?>>{};

  static Map<String, StreamController<UserPlatform?>> _userChangesListeners =
      <String, StreamController<UserPlatform?>>{};

  static Map<int, PhoneAuthCallbacks> _phoneAuthCallbacks = {};

  /// Returns a stub instance to allow the platform interface to access
  /// the class instance statically.
  static MethodChannelFirebaseAuth get instance {
    return MethodChannelFirebaseAuth._();
  }

  /// Internal stub class initializer.
  ///
  /// When the user code calls an auth method, the real instance is
  /// then initialized via the [delegateFor] method.
  MethodChannelFirebaseAuth._() : super(appInstance: null);

  /// Creates a new instance with a given [FirebaseApp].
  MethodChannelFirebaseAuth({required FirebaseApp app})
      : super(appInstance: app) {
    // Send a request to start listening to change listeners straight away
    channel
        .invokeMethod<void>('Auth#registerChangeListeners', <String, dynamic>{
      'appName': app.name,
    });

    // Create a app instance broadcast stream for native listener events
    _authStateChangesListeners[app.name] =
        StreamController<UserPlatform?>.broadcast();
    _idTokenChangesListeners[app.name] =
        StreamController<UserPlatform?>.broadcast();
    _userChangesListeners[app.name] =
        StreamController<UserPlatform?>.broadcast();

    // The channel setMethodCallHandler callback is not app specific, so there
    // is no need to register the caller more than once.
    if (_initialized) return;

    channel.setMethodCallHandler((MethodCall call) async {
      Map<dynamic, dynamic> arguments = call.arguments as Map<dynamic, dynamic>;

      switch (call.method) {
        case 'Auth#authStateChanges':
          // TODO(rrousselGit): find out why an exception here isn't reported to tests
          return _handleAuthStateChangesListener(arguments);
        case 'Auth#idTokenChanges':
          return _handleIdTokenChangesListener(arguments);
        case 'Auth#phoneVerificationCompleted':
          return _handlePhoneVerificationCompleted(arguments);
        case 'Auth#phoneVerificationFailed':
          return _handlePhoneVerificationFailed(arguments);
        case 'Auth#phoneCodeSent':
          return _handlePhoneCodeSent(arguments);
        case 'Auth#phoneCodeAutoRetrievalTimeout':
          return _handlePhoneCodeAutoRetrievalTimeout(arguments);
        default:
          throw UnimplementedError('${call.method} has not been implemented');
      }
    });

    _initialized = true;
  }

  @override
  UserPlatform? currentUser;

  @override
  String? languageCode;

  @override
  void sendAuthChangesEvent(String appName, UserPlatform? userPlatform) {
    assert(_userChangesListeners[appName] != null);

    _userChangesListeners[appName]!.add(userPlatform);
  }

  /// Handles any incoming [authChanges] listener events.
  // Duplicate setting of [currentUser] in [_handleAuthStateChangesListener] & [_handleIdTokenChangesListener]
  // as iOS & Android do not guarantee correct ordering
  Future<void> _handleAuthStateChangesListener(
    Map<dynamic, dynamic> arguments,
  ) async {
    String appName = arguments['appName'];
    // ignore: close_sinks, the sink is not created here, only obtained
    StreamController<UserPlatform?> streamController =
        _authStateChangesListeners[appName]!;
    MethodChannelFirebaseAuth instance =
        _methodChannelFirebaseAuthInstances[appName]!;

    final Map? userMap = arguments['user'] as Map?;

    if (userMap == null) {
      instance.currentUser = null;
      streamController.add(null);
    } else {
      final MethodChannelUser user =
          MethodChannelUser(instance, userMap.cast<String, dynamic>());

      // TODO(rousselGit): should this logic be moved to the setter instead?
      instance.currentUser = user;
      streamController.add(instance.currentUser);
    }
  }

  /// Handles any incoming [idTokenChanges] listener events.
  ///
  /// This handler also manages the [currentUser] along with sending events
  /// to any [userChanges] stream subscribers.
  Future<void> _handleIdTokenChangesListener(
    Map<dynamic, dynamic> arguments,
  ) async {
    String appName = arguments['appName'];
    // ignore: close_sinks, the sink is not created here, only obtained
    StreamController<UserPlatform?> idTokenStreamController =
        _idTokenChangesListeners[appName]!;
    // ignore: close_sinks, the sink is not created here, only obtained
    StreamController<UserPlatform?> userChangesStreamController =
        _userChangesListeners[appName]!;
    MethodChannelFirebaseAuth instance =
        _methodChannelFirebaseAuthInstances[appName]!;

    final Map? userMap = arguments['user'] as Map?;

    if (userMap == null) {
      instance.currentUser = null;
      idTokenStreamController.add(null);
      userChangesStreamController.add(null);
    } else {
      final MethodChannelUser user =
          MethodChannelUser(instance, userMap.cast<String, dynamic>());

      // TODO(rousselGit): should this logic be moved to the setter instead?
      instance.currentUser = user;
      idTokenStreamController.add(user);
      userChangesStreamController.add(user);
    }
  }

  Future<void> _handlePhoneVerificationCompleted(
    Map<dynamic, dynamic> arguments,
  ) async {
    final int handle = arguments['handle'];
    final int token = arguments['token'];
    final String smsCode = arguments['smsCode'];

    PhoneAuthCredential phoneAuthCredential =
        PhoneAuthProvider.credentialFromToken(token, smsCode: smsCode)
            as PhoneAuthCredential;
    PhoneAuthCallbacks callbacks = _phoneAuthCallbacks[handle]!;
    callbacks.verificationCompleted(phoneAuthCredential);
  }

  Future<void> _handlePhoneVerificationFailed(
    Map<dynamic, dynamic> arguments,
  ) async {
    final int handle = arguments['handle'];
    final Map<dynamic, dynamic> error = arguments['error'];
    final Map<dynamic, dynamic>? details = error['details'];

    PhoneAuthCallbacks callbacks = _phoneAuthCallbacks[handle]!;

    FirebaseAuthException exception = FirebaseAuthException(
      message: details != null ? details['message'] : error['message'],
      code: details != null ? details['code'] : 'unknown',
    );

    callbacks.verificationFailed(exception);
  }

  Future<void> _handlePhoneCodeSent(
    Map<dynamic, dynamic> arguments,
  ) async {
    final int handle = arguments['handle'];
    final String verificationId = arguments['verificationId'];
    final int? forceResendingToken = arguments['forceResendingToken'];

    PhoneAuthCallbacks callbacks = _phoneAuthCallbacks[handle]!;
    callbacks.codeSent(verificationId, forceResendingToken);
  }

  Future<void> _handlePhoneCodeAutoRetrievalTimeout(
    Map<dynamic, dynamic> arguments,
  ) async {
    final int handle = arguments['handle'];
    final String verificationId = arguments['verificationId'];

    PhoneAuthCallbacks callbacks = _phoneAuthCallbacks[handle]!;
    callbacks.codeAutoRetrievalTimeout(verificationId);
  }

  /// Gets a [FirebaseAuthPlatform] with specific arguments such as a different
  /// [FirebaseApp].
  ///
  /// Instances are cached and reused for incoming event handlers.
  @override
  FirebaseAuthPlatform delegateFor({required FirebaseApp app}) {
    return _methodChannelFirebaseAuthInstances.putIfAbsent(app.name, () {
      return MethodChannelFirebaseAuth(app: app);
    });
  }

  @override
  MethodChannelFirebaseAuth setInitialValues({
    Map<String, dynamic>? currentUser,
    String? languageCode,
  }) {
    if (currentUser != null) {
      this.currentUser = MethodChannelUser(this, currentUser);
    }

    this.languageCode = languageCode;
    return this;
  }

  @override
  Future<void> applyActionCode(String code) async {
    try {
      await channel
          .invokeMethod<void>('Auth#applyActionCode', <String, dynamic>{
        'appName': app.name,
        'code': code,
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<ActionCodeInfo> checkActionCode(String code) async {
    try {
      Map<String, dynamic> result = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#checkActionCode', <String, dynamic>{
        'appName': app.name,
        'code': code,
      }))!;

      return ActionCodeInfo(
        operation: result['operation'],
        data: Map<String, dynamic>.from(result['data']),
      );
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> confirmPasswordReset(String code, String newPassword) async {
    try {
      await channel
          .invokeMethod<void>('Auth#confirmPasswordReset', <String, dynamic>{
        'appName': app.name,
        'code': code,
        'newPassword': newPassword,
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> createUserWithEmailAndPassword(
      String email, String password) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#createUserWithEmailAndPassword', <String, dynamic>{
        'appName': app.name,
        'email': email,
        'password': password,
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<List<String>> fetchSignInMethodsForEmail(String email) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#fetchSignInMethodsForEmail', <String, dynamic>{
        'appName': app.name,
        'email': email,
      }))!;

      return List<String>.from(data['providers']);
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Stream<UserPlatform?> authStateChanges() =>
      _authStateChangesListeners[app.name]!.stream;

  @override
  Stream<UserPlatform?> idTokenChanges() =>
      _idTokenChangesListeners[app.name]!.stream;

  @override
  Stream<UserPlatform?> userChanges() =>
      _userChangesListeners[app.name]!.stream;

  @override
  Future<void> sendPasswordResetEmail(
    String email, [
    ActionCodeSettings? actionCodeSettings,
  ]) async {
    try {
      await channel
          .invokeMethod<void>('Auth#sendPasswordResetEmail', <String, dynamic>{
        'appName': app.name,
        'email': email,
        'actionCodeSettings': actionCodeSettings?.asMap(),
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> sendSignInLinkToEmail(
    String email,
    ActionCodeSettings actionCodeSettings,
  ) async {
    try {
      await channel
          .invokeMethod<void>('Auth#sendSignInLinkToEmail', <String, dynamic>{
        'appName': app.name,
        'email': email,
        'actionCodeSettings': actionCodeSettings.asMap(),
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> setLanguageCode(String languageCode) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#setLanguageCode', <String, dynamic>{
        'appName': app.name,
        'languageCode': languageCode,
      }))!;

      this.languageCode = data['languageCode'];
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> setSettings({
    bool? appVerificationDisabledForTesting,
    String? userAccessGroup,
  }) async {
    try {
      await channel.invokeMethod('Auth#setSettings', <String, dynamic>{
        'appName': app.name,
        'appVerificationDisabledForTesting': appVerificationDisabledForTesting,
        'userAccessGroup': userAccessGroup,
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> setPersistence(Persistence persistence) {
    throw UnimplementedError(
        'setPersistence() is only supported on web based platforms');
  }

  @override
  Future<UserCredentialPlatform> signInAnonymously() async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#signInAnonymously', <String, dynamic>{
        'appName': app.name,
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> signInWithCredential(
    AuthCredential credential,
  ) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#signInWithCredential', <String, dynamic>{
        'appName': app.name,
        'credential': credential.asMap(),
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> signInWithCustomToken(String token) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#signInWithCustomToken', <String, dynamic>{
        'appName': app.name,
        'token': token,
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> signInWithEmailAndPassword(
      String email, String password) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#signInWithEmailAndPassword', <String, dynamic>{
        'appName': app.name,
        'email': email,
        'password': password,
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> signInWithEmailLink(
      String email, String emailLink) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#signInWithEmailLink', <String, dynamic>{
        'appName': app.name,
        'email': email,
        'emailLink': emailLink,
      }))!;

      MethodChannelUserCredential userCredential =
          MethodChannelUserCredential(this, data);

      currentUser = userCredential.user;
      return userCredential;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<UserCredentialPlatform> signInWithPopup(AuthProvider provider) {
    throw UnimplementedError(
      'signInWithPopup() is only supported on web based platforms',
    );
  }

  @override
  Future<void> signInWithRedirect(AuthProvider provider) {
    throw UnimplementedError(
      'signInWithRedirect() is only supported on web based platforms',
    );
  }

  @override
  Future<void> signOut() async {
    try {
      await channel.invokeMethod<void>('Auth#signOut', <String, dynamic>{
        'appName': app.name,
      });

      currentUser = null;
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<String> verifyPasswordResetCode(String code) async {
    try {
      Map<String, dynamic> data = (await channel
          .invokeMapMethod<String, dynamic>(
              'Auth#verifyPasswordResetCode', <String, dynamic>{
        'appName': app.name,
        'code': code,
      }))!;

      return data['email'];
    } catch (e) {
      throw convertPlatformException(e);
    }
  }

  @override
  Future<void> verifyPhoneNumber({
    required String phoneNumber,
    required PhoneVerificationCompleted verificationCompleted,
    required PhoneVerificationFailed verificationFailed,
    required PhoneCodeSent codeSent,
    required PhoneCodeAutoRetrievalTimeout codeAutoRetrievalTimeout,
    String? autoRetrievedSmsCodeForTesting,
    Duration timeout = const Duration(seconds: 30),
    int? forceResendingToken,
  }) async {
    if (defaultTargetPlatform == TargetPlatform.macOS) {
      throw UnimplementedError(
        'verifyPhoneNumber() is not available on MacOS platforms.',
      );
    }

    int handle = MethodChannelFirebaseAuth.nextMethodChannelHandleId;

    _phoneAuthCallbacks[handle] = PhoneAuthCallbacks(verificationCompleted,
        verificationFailed, codeSent, codeAutoRetrievalTimeout);

    try {
      await channel
          .invokeMethod<void>('Auth#verifyPhoneNumber', <String, dynamic>{
        'appName': app.name,
        'handle': handle,
        'phoneNumber': phoneNumber,
        'timeout': timeout.inMilliseconds,
        'forceResendingToken': forceResendingToken,
        'autoRetrievedSmsCodeForTesting': autoRetrievedSmsCodeForTesting,
      });
    } catch (e) {
      throw convertPlatformException(e);
    }
  }
}
